/********************************************************************/
/*                                                                  */
/*   Program Name :  speana                                         */
/*                                                                  */
/*      Version:  2.10  2006-01-31                                  */
/*                                                                  */
/*      Copyright (c) 2002-2006 T.Kondo/NICT All Right Reserved     */
/*                                                                  */
/********************************************************************/
/*
D:\IPVLBI\data\k5vssp32\kas11ft32a.dat 21 1 0 K5/VSSP32 2
  Spectrum analysis of IP-VLBI board data
        (off-line processing)

          T.Kondo 2002.5.20
               Ver 0.00 2002.05.20 
               Ver 1.41 2002.10.01 18000->16000 bug fix
               Ver 1.42 2002.10.11 set default frequency axis linear
               Ver 1.43 2004.10.11 support sampling Frequency > 16MHz
               Ver 1.44 2005.06.12 update due to update in common header name
               Ver 1.45 2005.09.27 Allow to use PGDISP as an environmetal variable
                                   for PPGPLOT device
               Ver 2.00 2005.10.28 Improve processing speed drastically
                                   support K5/VSSP32 data processing
                                   FFTW3.0 is supported if hope so.
               Ver 2.10 2006.01.31 support linear power axis

    How to run:
       speana filename [mode [sekibun [pmode [comment [soffset]]]]]

        where filename -- data file name
                            0 means "tds.data" file
               mode -- YX: axis scale mode
                          X: frequency axis mode
                             =0: log scale  =1: linear scale (default)
                          Y: amplitude axis mode
                             =0: log scale (default)  
                             =1: linear scale (common for 4 channels)
                             =2: linear scale (each max and min)
                      -value: auto correlation (value point center)
                      -1:  auto correlation (full range)
                      
               sekibun  -- integration period (integer second)
                            default is 1 
               pmode -- plot device mode
                       0: XTERM and PostScript faile out (default)
                       1: PostScrip out only 
                       2: XTERM only 
               comment  -- comment (displayed at upper side of graph)
                           (do not include space)
                           If omitted, you will be asked later.
                           NOTE: In the interactive mode, you can sue space.
               soffset -- start offset (sec)
*/
#define SPEANA_VERSION "Ver. 2006-01-31"
#ifdef WIN32
  #define USE_FFTW    // use FFTW package
#endif

#ifdef USE_FFTW
 #ifdef WIN32
  #include "fftw3.h"     // use of FFTW package
 #else
  #include <fftw3.h>     // use of FFTW package
 #endif
#endif

#include "libk5.h"      // common header for IP-VLBI utilities

#include <cpgplot.h>
#include "libfft.h"

#ifdef WIN32
#define USE_ASM    // use inline assemble function
#endif

#include "libvssp32.h"

/*    Offset from dB to dBm conversion from A/D multibit sampling data */ 
#define OFFSET1  16.0     //1 bit sampling full power convert to 0 dB (theoretical)
      /* following constants are provisional  (measured on 2002/5/18) */
#define OFFSET2   9.0     //2 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET4  -1.0     //4 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET8 -26.0     //8 bit sampling data offset measured at 200kHz for PC(vlbi11)

#include "libspeplot.h"

int main(int argc, char *argv[])
{
    int sfreq, adbit, numch;
    long int startoffset; // start time offset in sec
    int nloop,sekibun;
    int i,imax, m, k, kaisu;
    int smode,usampl,nfft,kplot;
    int kword;
    int bits, bytes, numb;
    int amode;    // >0 auto correlation plot mode 
    int mode;  // frequency axis mode  0: log  1: linear 
    int pmode;
    int mode2;  // 2 digits axis mode
    int ymode; // amplitude scale mode   0: log   1: linear
    int scount[4];
    float fsampl,tsampl,fact;
    double *cdata;
    double *f_hanning;
    double theta;
    float *xdata, *ydata;
    double sum,dc;
    float pows, *spows;
    float pmax,pmin,pmaxdbm,pmindbm,cofst;
    float ymaxdb,ymindb;
#ifdef USE_FFTW
    fftw_complex *cc;
    fftw_plan pfcc, pbcc;
#endif
    double nfact;
    double telapsed;           // to measure processing time
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char fname[60];
    char lab[60], lab2[60], iplab[40], ltime1st[60];
    char labfsampl[10];
    char pgdevice[20];
    char* dflt_fname="tds.data";
    K5VSSP_file_t K5;

    /* set graphic device */
    strcpy(pgdevice,select_env_dflt(PGDISP, PGDISP_DFLT));

    if(argc == 1){
          printf("speana %s",SPEANA_VERSION);
#ifdef USE_FFTW
          printf(" compliled with FFTW3.0 \n");
#else
          printf("\n");
#endif
          printf("Usage: speana filename [mode [sekibun [pmode [comment [soffset]]]]]\n");
          printf("\n");
          printf("         where  filename ---- Data file name\n");
          printf("                              if 0 is given, tds.data will be used\n");
          printf("                mode -------- set axis mode or autocorrelation mode\n");
          printf("                              =0 : amplitude log, frequency log\n");
          printf("                              =1 (deflt) : amplitude log, frequency linear\n");
          printf("                              =10 : amplitude linear, frequency log\n");
          printf("                              =11 : amplitude linear, frequency linear\n");
          printf("                              =-1 : Auto correlation function plot\n");
          printf("                                     instead of spectrum\n");
          printf("                              =-N : N points auto correlation function plot\n");
          printf("                sekibun --- Averaging period (integer seconds)\n");
          printf("                             default is 1\n");
          printf("                pmode ---- Plot device selection\n");
          printf("                           =0 :  %s and PostScript file (pgplot.ps) out\n",pgdevice);
          printf("                           =1 :  PostScript file (pgplot.ps) out only\n");
          printf("                           =2 :  %s only\n",pgdevice);
          printf("                          default is 0\n");
          printf("                comment --- Comment appeared on a graph\n");
          printf("                            if omitted, you will be asked later.\n");
          printf("                soffset --- Start time offset in sec (integer seconds)\n");
          printf("                             default is 0\n"); 
          printf("\n");
          printf("  Note: %s ---  environment variable for PGPLOT display device\n",PGDISP);
          printf("                   = %s",getenv(PGDISP));
          if(getenv(PGDISP)==NULL) {
              printf(",  program deflt is ( %s )\n",PGDISP_DFLT);
          } else {
              printf("\n");
          }
          printf("\n");
          exit(0);
    } 

    strcpy(fname,*(argv+1));
    if (fname[0]=='0'){
          strcpy(fname,dflt_fname);
    }
    strcpy(K5.fname,fname);
    /* plot mode default set */
    mode2=1;
    mode=1;   // default frequency axis mode (=Linear sacle)
    ymode=0;  // default amplitude axis mode (=LOG scale)
    amode=0;  // auto correlation mode off
    /* plot mode set from operator */
    if (argc>=3){
       mode2 = atoi(*(argv + 2 ));
       if(mode2 <0 ){
            amode=-mode2;
            mode=0;
       } else {
            amode=0;
            ymode=mode2/10;
            mode=mode2 % 10;
            
       }
    }

    if (argc>=4){
       sekibun = atoi(*(argv + 3 ));
    } else {
          sekibun=1;     // default integration time (sec)
    }

    if (argc>=5){
       pmode = atoi(*(argv + 4 ));
    } else {
          pmode=0;     // default plot mode
    }

    printf("speana %s",SPEANA_VERSION);
#ifdef USE_FFTW
    printf(" compliled with FFTW3.0 \n");
#else
    printf("\n");
#endif

    cstart1sec = clock();

    printf("Data File is %s\n",fname);
    K5.size=get_file_size(K5.fname);
    if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",K5.fname);
        goto exitend2;
    }
    K5.twordread=0;  // reset word read counter
    K5.pos=0;        // reset current pointer position (in bytes)

    /* get comment */

    if (argc>=6){
          strcpy(iplab,*(argv+5));
    } else {
          /* enter board ID */
          /*
          printf("Enter IP-VLBI BOARD ID -> ");
          read1line(iplab);      // 1 line read from stdin
          printf("Board ID is %s\n",iplab);      
          strcpy(lab2,"IP-VLBI Board ID : ");
          strcat(lab2,iplab);
          */

          /* enter any comment */

          printf("Enter any comment -> ");
          read1line(iplab);      // 1 line read from stdin
          //printf("Board ID is %s\n",iplab);      
          //strcat(lab2,iplab);
    }
    if (argc>=7){
       startoffset = (long int)atoi(*(argv + 6 ));
    } else {
       startoffset=0;     // default startoffset
    }

    /* Data header check to get sampling parameters */
    if((kword=CheckK5header_with_Read(&K5)) < 0){
        printf("File read error! \n");
        //fgetpos(K5.stream,&K5.pos); 
        i=fclose( K5.stream ) ;
        goto exitend;
    }

//nextstep:
    /* set graphic device */
    strcpy(pgdevice,select_env_dflt(PGDISP, PGDISP_DFLT));

    /* parameter set */
    sfreq=K5.sfreq;
    adbit=K5.adbit;
    numch=K5.numch;
    fsampl=(float)(1000.0*(float)sfreq);   // sampling frequency (Hz)
    tsampl=(float)(1.0/fsampl);            // sampling period (sec)
    bits=((long)fsampl)*adbit*numch;  // # of bits in 1 sec data except header
    bytes=bits/8;                     // # of bytes in 1 sec data except header
    K5.fsample=fsampl;
    //sekibun=1;                       // Averaging period for Spectrum (sec)
    K5.bytesec=bits/8;                  //# of bytes in 1 sec data

    smode=1;                         // low resolution mode

    if(smode ==1){ 
        usampl=10000; 
    } else {
        usampl=50000;
        if(sfreq == 40)  usampl=40000; 
        if(sfreq == 100) usampl=100000; 
        if(sfreq == 200) usampl=200000; 
    }
    imax=(sfreq*1000)/usampl;
    numb=usampl*adbit*numch/8;   // # of bytes in unit # of samples
    K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
    K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second

    printf("imax numb  %d %d\n",K5.imax,K5.numb);    //debug

    nfft=1024;
    if(usampl < 524288) nfft= 524288;
    if(usampl < 262144) nfft= 262144;
    if(usampl < 131072) nfft= 131072;
    if(usampl < 65536)  nfft= 65536;
    if(usampl < 32768)  nfft= 32768;
    if(usampl < 16384)  nfft= 16384;

    nfact=2.0/(double)nfft;

    K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl
    K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

    K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
    K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

#ifdef USE_FFTW
    cc=fftw_malloc(sizeof(fftw_complex)*nfft);

    pfcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_FORWARD, FFTW_ESTIMATE);
    pbcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif

    //ibuf = (unsigned char *)malloc(numb);
    //idata = (int *)malloc(4*usampl*sizeof(int));
    cdata = (double *)malloc(2*nfft*sizeof(double));
    spows = (float *)malloc(4*nfft*sizeof(float));
    xdata = (float *)malloc(nfft*sizeof(float));
    ydata = (float *)malloc(nfft*sizeof(float));
    f_hanning = (double *)malloc(nfft*sizeof(double));
    //autocor=(float *)malloc(4*nfft*sizeof(float));

    //Hanning filter coeeficients prepare
    for(k=0; k<nfft; k++){
        theta=360.0*(double)k/(double)(nfft-1) - 180.0;
        theta=fmod(theta,360.0) ;
        f_hanning[k]=0.5*(1.0+cos(theta*PIRA));
    }

    fact=((float)usampl)*((float)usampl);
        
    /* sampling frequency in text for print */
    if(sfreq < 1000){
       sprintf(labfsampl,"%dkHz\0",sfreq);
    } else {
       sprintf(labfsampl,"%dMHz\0",sfreq/1000);
    }


    /* foward startffset seconds */
    if(fwdNsecK5(startoffset, &K5) <0){
        printf("Foward error %d\n",startoffset);
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    for(k=0; k<4*nfft; k++) spows[k]=0.0;
    for(k=0; k<4; k++) scount[k]=0;

    printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
        K5.hh,K5.mm,K5.ss,K5.seconds);
    sprintf(ltime1st,"Time(UT) %02d:%02d:%02d\0",K5.hh,K5.mm,K5.ss);
    strcpy(lab2,ltime1st);
    strcat(lab2,"  ");
    strcat(lab2,iplab);
        

    nloop=0;
loop1:
    if(nloop >0) {
        if((kword=CheckK5header_with_Read(&K5)) < 0){
            i=fclose( K5.stream ) ;
            goto exitend;
        }
        printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
            K5.hh,K5.mm,K5.ss,K5.seconds);
    }

    for(i=0; i<K5.imax; i++){
        K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
        K5.pos+=K5.wordread*4;
        K5.vflag=1; // validiate data
        parseADdata_3N(&K5);
        for(m=0; m<numch; m++){
            kaisu=0;
            sum=0.0;

            // initialize array
#ifdef USE_FFTW
            for(k=0; k<nfft; k++){ cc[k][0]=0.0; cc[k][1]=0.0;}
#else
            for(k=0; k<2*nfft; k++) cdata[k]=0.0;
#endif

            for(k=0; k<K5.npos/K5.numch; k++){

                /* set data */
#ifdef USE_FFTW
                cc[k][0]=(double)K5.idata[k*numch+m];  
                sum+=cc[k][0];
#else
                cdata[2*k]=(double)K5.idata[k*numch+m];  
                sum+=cdata[2*k];
#endif
                kaisu++;
            }
             
            dc=sum/(double)kaisu;    // DC component
     
            // DC offset removal from data
            for(k=0; k<K5.npos/numch; k++){
#ifdef USE_FFTW
                cc[k][0]-=dc;
#else
                cdata[2*k]-=dc;
#endif
            }
            //hanning(cdata,nfft);  // hanning filter
            //cdft(2*nfft,-1,cdata);
            if(amode ==0) {    // spectrum mode
                for(k=0; k<nfft; k++){
#ifdef USE_FFTW
                   cc[k][0]*=f_hanning[k];  // hanning filter
                   cc[k][1]*=f_hanning[k];  // hanning filter
#else
                   cdata[2*k]*=f_hanning[k];
                   cdata[2*k+1]*=f_hanning[k];
#endif
                }
                //hanning(cdata,usampl);  // hanning filter
#ifdef USE_FFTW
                fftw_execute(pbcc);
#else
                cdft(2*nfft,1,cdata);
#endif
            } else {
#ifdef USE_FFTW
                fftw_execute(pfcc);
                for(k=0; k<nfft; k++){
                    cc[k][0]*=nfact; cc[k][1]*=nfact;
                }
#else
                cdft(2*nfft,-1,cdata);
#endif
            }
            for(k=0; k<nfft; k++){
#ifdef USE_FFTW
                pows=(float)((cc[k][0]*cc[k][0]+cc[k][1]*cc[k][1])/(double)fact);
#else
                pows=(float)((cdata[2*k]*cdata[2*k]+cdata[2*k+1]*cdata[2*k+1])/(double)fact);
#endif
                spows[k*4+m]=spows[k*4+m]+pows;
            }

            scount[m]++;
        }
    }
    nloop++;
    if(nloop < sekibun) goto loop1;
    i=fclose( K5.stream ) ;
    //printf("%d bytes read\n",pos); 

    if(amode > 0) {   // auto correlation function calc
       for(m=0; m<numch; m++){
          for(k=0; k<nfft; k++){
             spows[k*4+m]/=scount[m]; // nomalization
#ifdef USE_FFTW
             cc[k][0]=spows[k*4+m]; cc[k][1]=0.0;
#else
             cdata[2*k]=spows[k*4+m];
             cdata[2*k+1]=0.0;
#endif
          }
          if(m==1 || m==3){
#ifdef USE_FFTW
              cc[0][0]=1.0;
#else
              cdata[0]=1.0;  // to avoid math error
#endif
          }
#ifdef USE_FFTW
          fftw_execute(pbcc);
#else
          cdft(2*nfft,1,cdata);
#endif
          for(k=0; k<nfft; k++){
             // if(k<10) printf("%d %e\n",k,cdata[2*k]);  // debug
             // if(k>nfft-10) printf("%d %e\n",k,cdata[2*k]);  // debug
#ifdef USE_FFTW
             if(k==0) pmax=(float)cc[0][0];
             if(cc[k][0] > pmax) pmax=(float)cc[k][0];
             spows[k*4+m]=(float)cc[k][0];
#else
             if(k==0) pmax=(float)cdata[0];
             if(cdata[2*k] > pmax) pmax=(float)cdata[2*k];
             spows[k*4+m]=(float)cdata[2*k];
#endif
          }
          for(k=0; k<nfft; k++){  // normalize
             spows[k*4+m]/=pmax;
             // if(k<10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
             // if(k>nfft-10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
          }
             
       }
         
    } else {

        /* convert power to dB  then dBm */       
        cofst=OFFSET1;       //1 bit sampling full power convert to 0 dB
        if( adbit==8){cofst=OFFSET8; }    // 8bit sampling offset
        if( adbit==4){cofst=OFFSET4; }    // 4bit sampling offset
        if( adbit==2){cofst=OFFSET2; }    // 4bit sampling offset
        for(m=0; m<numch; m++){
            for(k=0; k<nfft; k++){
                spows[k*4+m]/=scount[m]; // nomalization
                spows[k*4+m]=(float)(10.0*log10(spows[k*4+m]) + cofst); 
                if(k==0){
                    pmaxdbm=spows[k*4+m];
                    pmindbm=spows[k*4+m];
                    if(m==0){
                        ymaxdb=pmaxdbm;
                        ymindb=pmindbm;
                    }
                } else {
                    if(spows[k*4+m] > pmaxdbm) pmaxdbm=spows[k*4+m]; 
                    if(spows[k*4+m] < pmindbm) pmindbm=spows[k*4+m]; 
                }
            }
            pmax=pmaxdbm-cofst;
            pmin=pmindbm-cofst;
            printf("CH# %d: Maximum data (dB, dBm) is %f %f\n", m+1,pmax,pmaxdbm);      
            if(pmaxdbm > ymaxdb) ymaxdb=pmaxdbm;
            if(pmindbm < ymindb) ymindb=pmindbm;
        }
        printf("All CH: Maximum and Minimum data (dBm) are %f %f\n", ymaxdb, ymindb);      

    }
    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);

    /* plot spectrum (or auto cor) POSTSCRIPT OUT and XTERM*/

    for(kplot=0; kplot<2; kplot++){
       if(pmode==1 && kplot==1) goto bypass;
       if(pmode==2 && kplot==0) goto bypass;
       //if (cpgopen("?") < 1)
       if(kplot==0) {
          if (cpgopen("/PS") < 1) goto exitend;
       }
       if(kplot==1) {
          if (cpgopen(pgdevice) < 1) goto exitend;
       }

       cpgsubp(2,2);  // divide plot pannel into 2X2

       for(m=0; m<numch; m++){
           sprintf(lab,"CH#:%1d %1dbit %s sampling, Tavg(sec)=%3d\0",m+1,
                      adbit,labfsampl,sekibun);
           //printf("%s\n",lab);
           for(k=0; k<nfft; k++) {
                xdata[k]=spows[k*4+m];
           }
           if(amode > 0) {
                plotautocor(xdata, nfft, sfreq, adbit, lab, lab2, amode);
           } else {
                plotspe_a(xdata, nfft, sfreq, adbit, lab, lab2, mode2, ymaxdb,ymindb);
           }

       }
       /* panel undivided to date write */
       cpgsubp(1,1);
       /* date etc write */
       cpgiden();
       /* close the graphics device */
       cpgclos();
 bypass:;
    }

 exitend: ;
      
    free(K5.ubuf);
    free(K5.idata);
    free(cdata);
    free(spows);
    free(xdata);
    free(ydata);

#ifdef USE_FFTW
    fftw_destroy_plan(pfcc);
    fftw_destroy_plan(pbcc);
    fftw_free(cc);
#endif

 exitend2: ;

    return 0;
}
